/* Build with:
   gcc -o diff-manifest diff-manifest.c parse-manifest.c -Wall -O2
*/
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>

#include "manifest.h"
#include "utils.h"

/**************** needed blobs diff ************************/

static int
blob_added (FileInfo *info,
	    void *data)
{
  string_collector *c = data;
  
  if (info->sha1 != NULL &&
      strcmp (info->sha1, "empty") != 0)
    string_collector_add (c, info->sha1);

  return 1;
}

static int
blob_changed (FileInfo *info1,
	      FileInfo *info2,
	      void *data)
{
  string_collector *c = data;
  
  if (info2->sha1 != NULL &&
      (info1->sha1 == NULL ||
       strcmp (info1->sha1, info2->sha1) != 0) &&
      strcmp (info2->sha1, "empty") != 0)
    string_collector_add (c, info2->sha1);

  return 1;
}

/**************** changed blobs diff ************************/

static int
blob_changes_changed (FileInfo *info1,
		      FileInfo *info2,
		      void *data)
{
  string_collector *c = data;
  char *s;
  
  if (info1->sha1 != NULL &&
      info2->sha1 != NULL &&
      strcmp (info1->sha1, info2->sha1) != 0 &&
      strcmp (info2->sha1, "empty") != 0) {
    s = malloc (strlen (info1->sha1) + 4 + strlen (info2->sha1) + 1);
    strcpy (s, info1->sha1);
    strcat (s, " -> ");
    strcat (s, info2->sha1);
    string_collector_add (c, s);
    free (s);
  }

  return 1;
}


/**************** visual diff ************************/

struct diff_data {
  int n_added;
  int n_removed;
  int n_changed;
};

static int
diff_added (FileInfo *info,
	    void *_data)
{
  struct diff_data *data = _data;
  
  printf ("+ ");
  print_file_info (info, stdout);
  data->n_added++;

  return 1;
}

static int
diff_removed (FileInfo *info,
	      void *_data)
{
  struct diff_data *data = _data;
  
  printf ("- ");
  print_file_info (info, stdout);
  data->n_removed++;

  return 1;
}

static int
diff_changed (FileInfo *info1,
	      FileInfo *info2,
	      void *_data)
{
  struct diff_data *data = _data;
  
  printf ("/ ");
  print_file_info (info1, stdout);
  printf ("\\ ");
  print_file_info (info2, stdout);
  data->n_changed++;

  return 1;
}
  

static void
usage (void)
{
  fprintf (stderr, "Usage: diff-manifest [-b|-d] <manifest1> <manifest2>\n");
}


int
main(int argc, char *argv[])
{
  Manifest *m1, *m2;
  int arg;
  int do_blobs = 0;
  int do_blob_changes = 0;
  
  arg = 1;

  while (arg < argc) {
    if (strcmp (argv[arg], "-b") == 0) {
      do_blobs = 1;
    } else if (strcmp (argv[arg], "-c") == 0) {
      do_blob_changes = 1;
    } else {
      break;
    }
    arg++;
  }
		
  if (argc - arg != 2) {
    usage ();
    return 1;
  }
  
  m1 = parse_manifest (argv[arg++]);
  if (m1 == NULL)
    return 1;
  
  m2 = parse_manifest (argv[arg++]);
  if (m2 == NULL)
    return 1;
  
  if (do_blobs) {
    string_collector *l;
    int i;
    
    l = string_collector_new ();
    manifest_diff (m1, m2, 0, blob_added, NULL, blob_changed, l);
    
    for (i = 0; i < l->n_strings; i++)
      printf ("%s\n", l->strings[i].string);
    
    string_collector_free (l);
  } else if (do_blob_changes) {
    string_collector *l;
    int i;
    
    l = string_collector_new ();
    manifest_diff (m1, m2, 0, NULL, NULL, blob_changes_changed, l);
    
    for (i = 0; i < l->n_strings; i++)
      printf ("%s\n", l->strings[i].string);
    
    string_collector_free (l);
  } else {
    struct diff_data data = { 0, 0, 0};
    
    manifest_diff (m1, m2, 0, diff_added, diff_removed, diff_changed, &data);
    printf ("%d new files, %d removed files, %d changed files\n",
	    data.n_added, data.n_removed, data.n_changed);
  }

  return 0;
}
