#define _GNU_SOURCE

#include <stdio.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <fcntl.h>

#include "utils.h"
#include "manifest.h"

static int
compare_file_info_name (const void *_info1, const void *_info2)
{
  const FileInfo *info1, *info2;

  info1 = _info1;
  info2 = _info2;
  
  return strcmp (info1->name, info2->name);
}

void
destroy_file_info (FileInfo *info)
{
  if (info->name)
    free (info->name);
  if (info->symlink)
    free (info->symlink);
  if (info->sha1)
    free (info->sha1);
}

Manifest *
manifest_new (void)
{
  Manifest *m;

  m = calloc (1, sizeof (Manifest));
  if (m) {
    m->n_infos = 0;
    m->n_allocated = 32;
    m->infos = malloc (m->n_allocated * sizeof (FileInfo));
  }
  return m;
}

void
manifest_free (Manifest *m)
{
  int i;

  for (i = 0; i < m->n_infos; i++)
    destroy_file_info (&m->infos[i]);

  free (m->infos);
  free (m);
}

void
manifest_add_file (Manifest *m,
		   FileInfo *info)
{
  if (m->n_infos == m->n_allocated)
    {
      m->n_allocated = 2 * m->n_allocated;
      m->infos = realloc (m->infos, m->n_allocated * sizeof (FileInfo));
    }

  m->infos[m->n_infos++] = *info;
}

void
manifest_sort (Manifest *m)
{
  qsort (m->infos, m->n_infos, sizeof(FileInfo), compare_file_info_name);
}

int
file_info_equal (FileInfo *info1, FileInfo *info2)
{
  
  return
    strcmp (info1->name, info2->name) == 0 &&
    info1->type == info2->type &&
    info1->mode == info2->mode &&
    info1->uid == info2->uid &&
    info1->gid == info2->gid &&
    info1->dev == info2->dev &&
    safe_strcmp (info1->symlink, info2->symlink) == 0 &&
    safe_strcmp (info1->sha1, info2->sha1) == 0;
}

static char *
read_line (FILE *in)
{
  char *line, *tmp;
  char buffer[4096];
  char *res;
  size_t buf_len, line_len;
  int got_eol;

  line = NULL;

  while (1) {
    res = fgets(buffer, sizeof (buffer), in);
    if (res == NULL)
      return line;

    got_eol = 0;
    buf_len = strlen (buffer);
    if (buf_len > 0 && buffer[buf_len-1] == '\n') {
      buf_len--;
      buffer[buf_len] = 0;
      got_eol = 1;
    }
    if (line == NULL)
      line = strdup (buffer);
    else {
      line_len = strlen (line);
      tmp = malloc (buf_len + line_len + 1);
      strcpy (tmp, line);
      strcat (tmp, buffer);
      free (line);
      line = tmp;
    }

    if (got_eol)
      return line;
  }
}

FileType
get_type (char *str, char **end)
{
  FileType type;
  
  if (strncmp (str, "blob ", 5) == 0)
    {
      type = FILE_TYPE_BLOB;
      *end = str + 5;
    }
  else if (strncmp (str, "dir ", 4) == 0)
    {
      type = FILE_TYPE_DIR;
      *end = str + 4;
    }
  else if (strncmp (str, "char ", 5) == 0)
    {
      type = FILE_TYPE_CHAR;
      *end = str + 5;
    }
  else if (strncmp (str, "block ", 6) == 0)
    {
      type = FILE_TYPE_BLOCK;
      *end = str + 6;
    }
  else if (strncmp (str, "fifo ", 5) == 0)
    {
      type = FILE_TYPE_FIFO;
      *end = str + 5;
    }
  else if (strncmp (str, "link ", 5) == 0)
    {
      type = FILE_TYPE_LINK;
      *end = str + 5;
    }
  else if (strncmp (str, "socket ", 7) == 0)
    {
      type = FILE_TYPE_SOCKET;
      *end = str + 7;
    }
  else {
    type = 0;
    *end = NULL;
  }

  return type;
}

static char *
get_string (char *str, char **end)
{
  char *p;

  p = str;
  while (*p != ' ' && *p != 0)
    p++;

  *end = p;
  return strndup (str, p - str);
}

static char *
get_name (char *str, char **end)
{
  char *p;
  char *quote;
  char *name, *d;
  
  if (*str != '"')
    return get_string (str, end);

  p = str;
  p++;
  quote = strchr (p, '"');
  if (quote == NULL) {
    *end = NULL;
    return NULL;
  }

  name = malloc (quote - p + 1);
  d = name;

  while (p < quote)
    {
      unsigned char c = *p++;

      if (c == '\\')
	{
	  c = (unsigned char)(*p++ - '0') << 6;
	  c |= (unsigned char)(*p++ - '0') << 3;
	  c |= (unsigned char)(*p++ - '0') << 0;
	}
      
      *d++ = c;
    }
  *d = 0;
  
  *end = quote + 1;
  
  return name;
}

int
parse_manifest_line (char *line, FileInfo *info, int line_no)
{
  char *p;

  memset (info, 0, sizeof (*info));
  
  info->type = get_type (line, &p);
  if (p == NULL) {
    fprintf (stderr, "parse error on line %d, unknown type\n", line_no);
    return 0;
  }

  info->mode = strtol (p, &p, 8);
  
  /* Skip space */
  if (*p++ != ' ') {
    fprintf (stderr, "parse error on line %d, no space after mode\n", line_no);
    return 0;
  }

  info->uid = strtol (p, &p, 10);
  
  /* Skip colon */
  if (*p++ != ':') {
    fprintf (stderr, "parse error on line %d, no colon after uid\n", line_no);
    return 0;
  }
  
  info->gid = strtol (p, &p, 10);
  
  /* Skip space */
  if (*p++ != ' ') {
    fprintf (stderr, "parse error on line %d, no space after gid\n", line_no);
    return 0;
  }

  switch (info->type) {
  case FILE_TYPE_BLOB:
    info->sha1 = get_string (p, &p);
    
    /* Skip space */
    if (*p++ != ' ') {
      fprintf (stderr, "parse error on line %d\n", line_no);
      return 0;
    }
    
    break;
  case FILE_TYPE_LINK:
    info->symlink = get_name (p, &p);
    
    /* Skip space */
    if (*p++ != ' ') {
      fprintf (stderr, "parse error on line %d\n", line_no);
      return 0;
    }
    break;
  case FILE_TYPE_CHAR:
  case FILE_TYPE_BLOCK:
    info->dev = strtol (p, &p, 10);
    
    /* Skip space */
    if (*p++ != ' ') {
      fprintf (stderr, "parse error on line %d\n", line_no);
      return 0;
    }
    break;
  case FILE_TYPE_DIR:
  case FILE_TYPE_FIFO:
  case FILE_TYPE_SOCKET:
    break;
  }
  
  info->name = get_name (p, &p);

  return 1;
}


Manifest *
parse_manifest (char *path)
{
  FILE *file;
  char *line;
  FileInfo info;
  int line_no;
  Manifest *m;

  file = fopen (path, "r");
  if (file == NULL)
    {
      fprintf (stderr, "Error opening file %s\n", path);
      return NULL;
    }

  m = manifest_new ();
  
  line_no = 0;
  while (m != NULL &&
	 (line = read_line (file)) != NULL) {
    if (parse_manifest_line (line, &info, line_no++)) {
      manifest_add_file (m, &info);
    } else {
      destroy_file_info (&info);
      manifest_free (m);
      m = NULL;
    }
    free (line);
  }
  
  fclose (file);

  if (m)
    manifest_sort (m);
  
  return m;
}

int
manifest_diff (Manifest            *m1,
	       Manifest            *m2,
	       int                  reverse,
	       file_info1_callback  added,
	       file_info1_callback  removed,
	       file_info2_callback  changed,
	       void                *data)
{
  int i1, i2;
  FileInfo *info1, *info2;

  if (reverse) {
    i1 = m1->n_infos - 1;
    i2 = m2->n_infos - 1;

    while (i1 >= 0 || i2 >= 0) {
      if (i1 >= 0)
	info1 = &m1->infos[i1];
      else
	info1 = NULL;
      
      if (i2 >= 0)
	info2 = &m2->infos[i2];
      else
	info2 = NULL;
      
      if (info1 == NULL ||
	  (info2 != NULL && strcmp (info1->name, info2->name) < 0)) {
	if (added) {
	  if (!added (info2, data))
	    return 0;
	}
	
	i2--;
      } else if (info2 == NULL ||
		 (info1 != NULL && strcmp (info1->name, info2->name) > 0)) {
	if (removed) {
	  if (!removed (info1, data))
	    return 0;
	}
	
	i1--;
      } else {
	assert (strcmp (info1->name, info2->name) == 0);
	
	if (changed && !file_info_equal (info1, info2)) {
	  if (!changed (info1, info2, data))
	    return 0;
	}
	
	i1--;
	i2--;
      }
    }
  } else /* forward */ {
    i1 = 0;
    i2 = 0;
    while ((i1 < m1->n_infos) || (i2 < m2->n_infos)) {
      if (i1 < m1->n_infos)
	info1 = &m1->infos[i1];
      else
	info1 = NULL;
      
      if (i2 < m2->n_infos)
	info2 = &m2->infos[i2];
      else
	info2 = NULL;
      
      if (info1 == NULL ||
	  (info2 != NULL && strcmp (info1->name, info2->name) > 0)) {
	if (added) {
	  if (!added (info2, data))
	    return 0;
	}
	
	i2++;
      } else if (info2 == NULL ||
		 (info1 != NULL && strcmp (info1->name, info2->name) < 0)) {
	if (removed) {
	  if (!removed (info1, data))
	    return 0;
	}
	
	i1++;
      } else {
	assert (strcmp (info1->name, info2->name) == 0);
	
	if (changed && !file_info_equal (info1, info2)) {
	  if (!changed (info1, info2, data))
	    return 0;
	}
	
	i1++;
	i2++;
      }
    }
  }
  return 1;
}

static int
should_escape (unsigned char c)
{
  if (c <= ' ' || c >= 127)
    return 1;

  return strchr ("!`\"$\\", c) != NULL;
}

static char *
escape_name (char *name)
{
  char *escaped, *d, *s;
  int n_escaped;

  n_escaped = 0;
  s = name;
  while (*s)
    {
      unsigned char c = *s++;
      if (should_escape (c))
	n_escaped++;
    }

  if (n_escaped == 0 && *name != 0)
    return strdup (name);
  
  escaped = malloc (strlen (name) + 3 * n_escaped + 3);
  d = escaped;
  s = name;

  *d++ = '"';
  while (*s)
    {
      unsigned char c = *s++;
      if (should_escape (c))
	{
	  *d++ = '\\';
	  *d++ = ((c >> 6) & ((1<<3)-1)) + '0';
	  *d++ = ((c >> 3) & ((1<<3)-1)) + '0';
	  *d++ = ((c >> 0) & ((1<<3)-1)) + '0';
	}
      else
	*d++ = c;
    }
  *d++ = '"';
  *d = 0;

  return escaped;
}

void
print_file_info (FileInfo *info, FILE *out)
{
  char *type;
  char *data;
  char *name;

  data = NULL;
  switch (info->type) {
  case FILE_TYPE_BLOB:
    type = "blob";
    data = strdup (info->sha1);
    break;
  case FILE_TYPE_DIR:
    type = "dir";
    break;
  case FILE_TYPE_CHAR:
    type = "char";
    data = malloc (20);
    snprintf (data, 20, "%u", (unsigned int) info->dev);
    break;
  case FILE_TYPE_BLOCK:
    type = "block";
    data = malloc (20);
    snprintf (data, 20, "%u", (unsigned int) info->dev);
    break;
  case FILE_TYPE_FIFO:
    type = "fifo";
    break;
  case FILE_TYPE_LINK:
    type = "link";
    data = escape_name (info->symlink);
    break;
  case FILE_TYPE_SOCKET:
    type = "socket";
    break;
  default:
    fprintf (stderr, "Broken type %d for %s, ignoring\n", info->type, info->name);
    /* Broken type, ignore */
    return;
  }

  name = escape_name (info->name);
  
  fprintf (out, "%s %o %d:%d ",
	   type,
	   info->mode,
	   info->uid,
	   info->gid);
  if (data)
    fprintf (out, "%s ", data);
  fprintf (out, "%s\n", name);
  
  free (name);

  if (data)
    free (data);
}

void
manifest_print (Manifest *m,
		FILE *out)
{
  int i;

  for (i = 0; i < m->n_infos; i++)
    print_file_info (&m->infos[i], out);
}

FileType
get_file_type_from_stat (struct stat *buf)
{
  
  if (S_ISREG(buf->st_mode))
    return FILE_TYPE_BLOB;
  else if (S_ISDIR(buf->st_mode))
    return FILE_TYPE_DIR;
  else if (S_ISCHR(buf->st_mode))
    return FILE_TYPE_CHAR;
  else if (S_ISBLK(buf->st_mode))
    return FILE_TYPE_BLOCK;
  else if (S_ISFIFO(buf->st_mode))
    return FILE_TYPE_FIFO;
  else if (S_ISLNK(buf->st_mode))
    return FILE_TYPE_LINK;
  else if (S_ISSOCK(buf->st_mode))
    return FILE_TYPE_SOCKET;
  else 
    return -1;
}

static int
generate_manifest_helper (Manifest *m, char *path, char *relative, int is_root)
{
  struct stat buf;
  int res;
  char *data;
  char *name;
  FileInfo info;

  res = lstat (path, &buf);
  if (res == -1)
    return 0;

  memset (&info, 0, sizeof(info));
  
  data = NULL;
  info.type = get_file_type_from_stat (&buf);
  if (info.type == -1) {
    fprintf (stderr, "unknown type for file %s, ignoring\n", path);
    return 0;
  }

  if (info.type == FILE_TYPE_BLOB) {
    if (buf.st_size == 0)
      info.sha1 = strdup ("empty");
    else {
      info.sha1 = sha1_file (path);
    }
  } else if (info.type == FILE_TYPE_CHAR ||
	     info.type == FILE_TYPE_BLOCK) {
    info.dev = buf.st_rdev;
  } else if (info.type == FILE_TYPE_LINK) {
    info.symlink = read_symlink (path);
    if (info.symlink == NULL) {
      fprintf (stderr, "error reading symlink value for file %s, ignoring\n", path);
      return 0;
    }
  }

  name = path + strlen (relative);
  while (*name == '/')
    name++;
  if (!is_root) /* Don't print root */
    {
      info.name = strdup (name);
      info.mode = buf.st_mode & ~S_IFMT;
      info.uid = buf.st_uid;
      info.gid = buf.st_gid;
      
      manifest_add_file (m, &info);
    }
  else
    destroy_file_info (&info);

  if (S_ISDIR (buf.st_mode)) {
    DIR *dir;
    struct dirent *dirent;
    char *child;
    
    dir = opendir (path);
    if (dir != NULL) {
      while ((dirent = readdir(dir)) != NULL) {
	if (strcmp (dirent->d_name, ".") == 0 ||
	    strcmp (dirent->d_name, "..") == 0)
	  continue;
	
	child = malloc (strlen (path) + strlen (dirent->d_name) + 2);
	strcpy (child, path);
	strcat (child, "/");
	strcat (child, dirent->d_name);
	
	if (!generate_manifest_helper (m, child, relative, 0))
	  return 0;
	free (child);
      }
      closedir (dir);
    }
  }
  
  return 1;
}

Manifest *
generate_manifest (char *path)
{
  Manifest *m;

  m = manifest_new ();

  if (!generate_manifest_helper (m, path, path, 1)) {
    fprintf (stderr, "error while processing %s\n", path);
    return NULL;
  }
  
  manifest_sort (m);
  return m;
}

int
file_info_create (FileInfo *info, const char *image_dir, const char *blob_dir, int blob_fd)
{
  char *file, *source;
  int res;
  int fd;
  struct stat buf;
  int already_dir;

  file = build_path (image_dir, info->name);

  already_dir = 0;
  /* Check for existing files */
  if (stat (file, &buf) == 0) {
    if (S_ISDIR(buf.st_mode)) {
	if (info->type != FILE_TYPE_DIR) {
	  res = rmdir(file);
	  if (res != 0) {
	    fprintf (stderr, "Unexpected non-empty directory in the way at %s\n", file);
	    free (file);
	    return 0;
	  }
	} else {
	  already_dir = 1;
	}
    } else {
      /* Not a directory */
      res = unlink(file);
      if (res != 0) {
	fprintf (stderr, "Unexpected non-removable file in the way at %s\n", file);
	free (file);
	return 0;
      }
    }
  }
 
  switch (info->type) {
  case FILE_TYPE_BLOB:
    if (strcmp (info->sha1, "empty") == 0) {
      fd = open (file, O_CREAT | O_WRONLY, info->mode);
      if (fd != -1) {
	close (fd);
      } else {
	fprintf (stderr, "Error creating empty file %s\n", file);
	free (file);
	return 0;
      }
    } else {
      if (blob_fd != -1) {
	res = copy_fd (blob_fd, file, info->mode);
      } else {
	source = build_path (blob_dir, info->sha1);
	res = copy_file (source, file, info->mode);
	free (source);
      }

      if (!res) {
	fprintf (stderr, "Error creating file %s\n", file);
	free (file);
	return 0;
      }
    }
    break;
    
  case FILE_TYPE_DIR:
    if (!already_dir) {
      res = mkdir (file, info->mode);
      if (res == -1) {
	fprintf (stderr, "Error creating directory %s\n", file);
	free (file);
	return 0;
      }
    }
    break;
    
  case FILE_TYPE_CHAR:
    res = mknod(file, info->mode | S_IFCHR, info->dev);
    if (res == -1) {
      fprintf (stderr, "Error creating char device %s\n", file);
      free (file);
      return 0;
    }
    break;
    
  case FILE_TYPE_BLOCK:
    res = mknod(file, info->mode | S_IFBLK, info->dev);
    if (res == -1) {
      fprintf (stderr, "Error creating char device %s\n", file);
      free (file);
      return 0;
    }
    break;
    
  case FILE_TYPE_FIFO:
    res = mknod(file, info->mode | S_IFIFO, info->dev);
    if (res == -1) {
      fprintf (stderr, "Error creating fifo %s\n", file);
      free (file);
      return 0;
    }
    break;
    
  case FILE_TYPE_LINK:
    res = symlink(info->symlink, file);
    if (res == -1) {
      fprintf (stderr, "Error creating symlink %s\n", file);
      free (file);
      return 0;
    }
    break;
    
  case FILE_TYPE_SOCKET:
    res = mknod(file, info->mode | S_IFSOCK, info->dev);
    if (res == -1) {
      fprintf (stderr, "Error creating socket %s\n", file);
      free (file);
      return 0;
    }
    break;
  }  

  /* TODO: We might be setting a mode on dirs so we can't modify them... needs to be
     done after handling children. Or we could just ignore this as we run the update as root
   */
	
  lchown (file, info->uid, info->gid);
  
  if (info->type != FILE_TYPE_LINK) { /* symlinks have no own mode */
    /* We have to do this with chmod, since umask affects the permission for new
     files, and the lchown call clears e.g. the setuid bit.  */
    res = chmod (file, info->mode);
    if (res == -1) {
      fprintf (stderr, "Error chmod:ing file %s\n", file);
      free (file);
      return 0;
    }
  }

  return 1;
}
