/* Build with:
   gcc -o diff-manifest diff-manifest.c parse-manifest.c -Wall -O2
*/
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include <avahi-client/client.h>
#include <avahi-client/publish.h>
#include <avahi-common/simple-watch.h>
#include <avahi-common/malloc.h>
#include <avahi-common/error.h>
#include <avahi-common/timeval.h>

#include "manifest.h"
#include "utils.h"

static void register_service (AvahiClient *c);

static AvahiSimplePoll *simple_poll = NULL;
static AvahiClient *client = NULL;
static AvahiEntryGroup *entry_group = NULL;
static AvahiStringList *txt_record = NULL;
static char *manifest_path = NULL;
static Manifest *manifest;
static char *version = NULL;
static char *image_id = NULL;
static char *image_dir;
static char *diff_dir = NULL;
static int port;

static int
create_server_socket (int *port_out) {
  struct sockaddr_in sin;
  int s;
  int val;
  socklen_t len;

  memset ((char *)&sin, 0, sizeof (sin));
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = INADDR_ANY;
  sin.sin_port = 0;
 
  /* Create a socket: */
  s = socket (PF_INET, SOCK_STREAM, 0);
  if (s < 0) {
    fprintf (stderr, "Error: couldn't create socket!\n");
    return -1;
  }

  val = 1;
  setsockopt (s, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val));

  /* bind the socket: */
  if (bind (s,(struct sockaddr *)&sin, sizeof (sin)) < 0) {
    fprintf (stderr, "Error: couldn't bind socket to port\n");
    return -1;
  }

  len = sizeof (sin);
  if (getsockname (s, (struct sockaddr *)&sin, &len) == -1) {
    fprintf (stderr, "Error: couldn't get socket port\n");
    return -1;
  }
  
  *port_out = ntohs (sin.sin_port);
  
  if (listen (s, 10) < 0) {
    fprintf (stderr, "Error: couldn't listen at port\n");
    return -1;
  }
  
  return s;
}

static int
write_file (const char *path, int fd_out)
{
  int fd_in;
  char buffer[4096];
  ssize_t res;
  size_t to_write;
  int pos;
  
  fd_in = open (path, O_RDONLY);
  if (fd_in == -1)
    return 0;
  
  while (1) {
    res = read (fd_in, buffer, sizeof(buffer));
    if (res == -1) {
      if (errno == EINTR)
	continue;
      close (fd_out);
      close (fd_in);
      return 0;
    }
    if (res == 0)
      break;

    to_write = res;
    pos = 0;

    while (pos < to_write) {
      res = write (fd_out, &buffer[pos], to_write - pos);
      if (res == -1) {
	if (errno == EINTR)
	  continue;
	close (fd_out);
	close (fd_in);
	return 0;
      }
      if (res == 0)
	{
	  /* Short write?? */
	  close (fd_out);
	  close (fd_in);
	  return 0;
	}
      pos += res;
    }
  }
  
  close (fd_in);
  
  return 1;
}

#define HTTP_ERROR_NOT_FOUND "HTTP/1.0 404 Not Found\r\n"
#define HTTP_ERROR_INTERNAL "HTTP/1.0 500 Internal Server Error\r\n"
#define HTTP_OK_W_LEN "HTTP/1.0 200 OK\r\nContent-Length: %lu\r\n\r\n"

static const char *
lookup_sha1 (const char *sha1)
{
  int i;
  
  for (i = 0; i < manifest->n_infos; i++)
    {
      if (manifest->infos[i].sha1 != NULL &&
	  strcmp (manifest->infos[i].sha1, sha1) == 0)
	return manifest->infos[i].name;
    }
  
  return NULL;
}

static int
str_has_prefix (const char *str, const char *prefix)
{
  int prefix_len;
  prefix_len = strlen (prefix);
  return strncmp (str, prefix, prefix_len) == 0;
}

static void
http_handle_request (int request_socket)
{
  char buffer[4096];
  char *filename, *p, *path;
  char *diff_file;
  const char *name;
  struct stat statbuf;
  int res;
  
  res = read (request_socket, buffer, sizeof (buffer) - 1); /* - 1 for null termination below */

  if (res == -1 ||
      res < 5 ||
      strncasecmp ("get /", buffer, 5) != 0)
    {
      write (request_socket, HTTP_ERROR_INTERNAL, strlen (HTTP_ERROR_INTERNAL));
      return;
    }
  
  p = filename = &buffer[5];
  while (p - buffer < res && *p != 0 && *p != ' ' && *p != '\n' && *p != '\r')
    p++;
  
  *p = 0;

  path = NULL;

  if (str_has_prefix (filename, "blobs/")) {
    name = lookup_sha1 (filename + strlen ("blobs/"));
    if (name != NULL)
      path = build_path (image_dir, name);
  } else if (str_has_prefix (filename, "manifests/") &&
	     str_has_prefix (filename + strlen ("manifests/"), image_id) &&
	     str_has_prefix (filename + strlen ("manifests/") + strlen (image_id) , ".") &&
	     strcmp (filename + strlen ("manifests/") + strlen (image_id) + 1 , version) == 0)
    path = strdup (manifest_path);
  else if (diff_dir != NULL && str_has_prefix (filename, "diffs/")) {
    diff_file = filename + strlen ("diffs/");
    if (strlen (diff_file) == 86 && diff_file[40] == '-' &&
	strcmp (diff_file + 40 + 1 + 40, ".diff") == 0)
      path = build_path (diff_dir, diff_file);
  }

  if (path == NULL)
    write (request_socket, HTTP_ERROR_NOT_FOUND, strlen (HTTP_ERROR_NOT_FOUND));
  else {
    res = stat (path, &statbuf);
    if (res == -1)
      write (request_socket, HTTP_ERROR_NOT_FOUND, strlen (HTTP_ERROR_NOT_FOUND));
    else
      {
	char result_buffer[1024];
	snprintf (result_buffer, 1024, HTTP_OK_W_LEN, (unsigned long)statbuf.st_size);
	write (request_socket, result_buffer, strlen (result_buffer));
	write_file (path, request_socket);
      }
    free (path);
  }
}

static void
server_socket_watch (AvahiWatch *w, int server_socket, AvahiWatchEvent event, void *userdata)

{
  int request_socket;
  struct sockaddr_in fsin;
  socklen_t alen;
  int val;
  
  alen = sizeof (fsin);
  request_socket = accept (server_socket, (struct sockaddr *)&fsin, &alen);
      
  if (request_socket != -1) {
    val = 1;
    setsockopt (request_socket, SOL_SOCKET, SO_KEEPALIVE, &val, sizeof (val));
    
    http_handle_request (request_socket);
    
    close (request_socket);
  }
}


static void
entry_group_callback (AvahiEntryGroup *g, AvahiEntryGroupState state, void *userdata)
{
  switch (state) {
  case AVAHI_ENTRY_GROUP_ESTABLISHED:
    break;

  case AVAHI_ENTRY_GROUP_FAILURE:
    fprintf(stderr, "Failed to register: %s\n", avahi_strerror (avahi_client_errno (client)));
    break;

  case AVAHI_ENTRY_GROUP_COLLISION:
    fprintf(stderr, "Name collision, picking new name.\n");
    register_service (client);
    break;
    
  case AVAHI_ENTRY_GROUP_UNCOMMITED:
  case AVAHI_ENTRY_GROUP_REGISTERING:
    break;
  }
}

static void
register_service (AvahiClient *c)
{
  char name[200], *p;
  int i;
  
  if (!entry_group) {
    entry_group = avahi_entry_group_new (c, entry_group_callback, NULL);
    
    if (entry_group == NULL) {
      fprintf(stderr, "Failed to create entry group: %s\n", avahi_strerror (avahi_client_errno (c)));
      return;
    }
  }

  strcpy (name, "laptop-");
  p = name + 7;
  for (i = 0; i < 8; i++)
    *p++ = "0123456789ABCDEF"[rand () % 16];
  *p = 0;

  if (avahi_entry_group_add_service_strlst (entry_group, AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC, 0,
					    name, "_updatinator._tcp", NULL, NULL, port,
					    txt_record)) {
    fprintf(stderr, "Failed to add service: %s\n", avahi_strerror (avahi_client_errno (c)));
    return;
  }
  
  avahi_entry_group_commit (entry_group);
  return;
}

static void
client_callback (AvahiClient *c, AvahiClientState state, void * userdata)
{
  int error;
  
  switch (state) {
  case AVAHI_CLIENT_S_RUNNING:
    register_service (c);
    break;
    
  case AVAHI_CLIENT_FAILURE:
    if (avahi_client_errno(c) == AVAHI_ERR_DISCONNECTED) {
      fprintf(stderr, "Disconnected, reconnecting ...\n");
      
      avahi_client_free (client);
      client = NULL;
      entry_group = NULL;
      
      client = avahi_client_new (avahi_simple_poll_get (simple_poll),
				 AVAHI_CLIENT_NO_FAIL, client_callback, NULL, &error);

      if (client == NULL) {
	fprintf (stderr, "Client failure, exiting: %s\n", avahi_strerror (error));
	avahi_simple_poll_quit (simple_poll);
      }
    } else {
      fprintf (stderr, "Client failure, exiting: %s\n", avahi_strerror (avahi_client_errno (c)));
      avahi_simple_poll_quit (simple_poll);
    }
    break;

  case AVAHI_CLIENT_S_COLLISION:
    
    /* Fall through */
    
  case AVAHI_CLIENT_S_REGISTERING:
    
    if (entry_group) {
      avahi_entry_group_free(entry_group);
      entry_group = NULL;
    }
    break;
    
  case AVAHI_CLIENT_CONNECTING:
    break;
  }
}

static void
usage (void)
{
  fprintf (stderr, "Usage: serve-manifest <manifest> <image dir> <image id> <version> [diff dir]\n");
}

int
main (int argc, char *argv[])
{
  int error;
  int arg;
  int server_socket;

  srand (time (NULL));
  
  arg = 1;

  if (argc - arg < 4) {
    usage ();
    return 1;
  }

  manifest_path = argv[arg++];
  manifest = parse_manifest (manifest_path);
  if (manifest == NULL)
    return 1;
  
  image_dir = argv[arg++];

  server_socket = create_server_socket (&port);
  if (server_socket == -1)
    return 1;

  image_id = argv[arg++];
  version = argv[arg++];

  if (argc - arg > 0)
    diff_dir = argv[arg++];

  txt_record = avahi_string_list_add_pair (NULL, "image", image_id);
  txt_record = avahi_string_list_add_pair (txt_record, "versions", version);
  txt_record = avahi_string_list_add_pair (txt_record, "prio", "1");
  if (diff_dir != NULL)
    txt_record = avahi_string_list_add (txt_record, "diff");

  if (!(simple_poll = avahi_simple_poll_new ())) {
    fprintf (stderr, "Failed to create simple poll object.\n");
    return 1;
  }

  client = avahi_client_new (avahi_simple_poll_get (simple_poll),
			     AVAHI_CLIENT_NO_FAIL, client_callback, NULL, &error);

  if (client == NULL) {
    fprintf (stderr, "Client failure, exiting: %s\n",
	     avahi_strerror (error));
    return 1;
  }
  
  /* Watch for connections to server socket */
  avahi_simple_poll_get (simple_poll)->watch_new (avahi_simple_poll_get (simple_poll),
						  server_socket,
						  AVAHI_WATCH_IN,
						  server_socket_watch,
						  NULL);

  avahi_simple_poll_loop(simple_poll);
  
  return 1; /* always return 1, we should never exit successfully */
}
