/* Build with:
   gcc -o diff-manifest diff-manifest.c parse-manifest.c -Wall -O2
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "manifest.h"
#include "utils.h"

static char *blob_dir = NULL;
static char *image_dir = NULL;
static int do_delete = 0;
static string_collector *blobs = NULL;

static int 
remove_file (FileInfo *info)
{
  char *file;
  int res;

  file = build_path (image_dir, info->name);
  
  if (info->type == FILE_TYPE_DIR)
    res = rmdir (file);
  else
    res = unlink (file);

  /* If its already removed, ignore failure */
  if (res == -1 && errno == ENOENT)
    res = 0;
    
  if (res != 0)
    fprintf (stderr, "Unable to remove file %s\n", file);
  
  free (file);

  return res == 0;
}

/* Collect blobs needed so we can move behind without having problems if one
   blob is used many times */

static int
blob_added (FileInfo *info,
	    void *data)
{
  string_collector *blobs = data;
  
  if (info->type == FILE_TYPE_BLOB &&
      strcmp (info->sha1, "empty") != 0)
    string_collector_add (blobs, info->sha1);

  return 1;
}

static int
blob_changed (FileInfo *info1,
	      FileInfo *info2,
	      void *data)
{
  string_collector *blobs = data;
  
  if (info2->type == FILE_TYPE_BLOB &&
      strcmp (info2->sha1, "empty") != 0 &&
      (info1->type != FILE_TYPE_BLOB ||
       strcmp (info1->sha1, info2->sha1) != 0))
    string_collector_add (blobs, info2->sha1);
  
  return 1;
}

static void
used_blob(char *sha1)
{
  string_info *info;
  char *path;

  if (!do_delete)
    return;
  
  if (strcmp(sha1, "empty") == 0)
    return;
  
  info = string_collector_lookup (blobs, sha1);

  if (info == NULL)
    return;

  if (info->count > 0)
    info->count--;
  
  if (info->count == 0) {
    path = build_path (blob_dir, sha1);
    unlink (path);
    free (path);
  }
}

/* First phase (forward) when creating files. This does all the adds, which means
   we create directories before the files in them */

static int
file1_added (FileInfo *info,
	     void *data)
{
  int res;

  res = file_info_create (info, image_dir, blob_dir, -1);

  if (info->type == FILE_TYPE_BLOB)
    used_blob(info->sha1);
  
  return res;
}

static int
file1_removed (FileInfo *info,
		 void *data)
{
  /* Removes are in the second phase */
  return 1;
}

static int
file1_changed (FileInfo *info1,
	       FileInfo *info2,
	       void *data)
{
  char *file;
  int res;
  int changed_owner;
  int blob_fd;
  
  if (info1->type == FILE_TYPE_DIR &&
      info2->type == FILE_TYPE_DIR)
    {
      /* We can't change dirs by remove + recreate, because they
	 can have files in them */
      
      file = build_path (image_dir, info2->name);

      /* Change owner before mode, since otherwise setuid bits, etc are
	 reset */
      changed_owner = 0;
      if (info1->uid != info2->uid ||
	  info1->gid != info2->gid) {
	changed_owner = 1;
	res = lchown (file, info2->uid, info2->gid);
	if (res == -1) {
	  fprintf (stderr, "Unable to change ownership of dir %s\n", file);
	  free (file);
	  return 0;
	}
      }

      /* Make sure we chmod if owner changed, to handle setuid bits reset */
      if (info1->mode != info2->mode || changed_owner) {
	res = chmod (file, info2->mode);
	if (res == -1) {
	  fprintf (stderr, "Unable to change mode of dir %s\n", file);
	  free (file);
	  return 0;
	}
      }
	
      free (file);

      return 1;
    }

  /* TODO: What about the case directory -> other type, and the directory is
     not empty. We should do a recursive remove in that case... or? */

  blob_fd = -1;

  if (info1->type == FILE_TYPE_BLOB && info2->type == FILE_TYPE_BLOB &&
      strcmp (info1->sha1, info2->sha1) == 0) {
    /* We want to use this data for the blob. Keep it open */
    file = build_path (image_dir, info2->name);
    blob_fd = open(file, O_RDONLY);
    if (blob_fd == -1) {
	  fprintf (stderr, "Unable to open required existing file %s\n", file);
	  free (file);
	  return 0;
    }
    free (file);
  }
  
  /* Recreate file, first removing it. This ensures that we
     break any links for files that are hardlinks, so that we don't
     change files outside the target file/image */
  if (!remove_file (info1))
    return 0;
  if (!file_info_create (info2, image_dir, blob_dir, blob_fd))
    return 0;

  if (blob_fd != -1)
    close (blob_fd);
  
  if (info2->type == FILE_TYPE_BLOB &&
      strcmp (info2->sha1, "empty") != 0 &&
      (info1->type != FILE_TYPE_BLOB ||
       strcmp (info1->sha1, info2->sha1) != 0))
    used_blob(info2->sha1);

  return 1;
}


/* Second phase, done backwards, when creating files. This does all the removes,
   which means remove files before the directories they are in */

static int
file2_added (FileInfo *info,
	       void *data)
{
  return 1;
}

static int
file2_removed (FileInfo *info,
		 void *data)
{
  return remove_file (info);
}

static int
file2_changed (FileInfo *info1,
		 FileInfo *info2,
		 void *data)
{
  return 1;
}

static void
usage (void)
{
  fprintf (stderr, "Usage: upgrade-manifest -b <blob_dir> -d <image_dir> [--delete] <manifest1> <manifest2>\n");
}


int
main(int argc, char *argv[])
{
  Manifest *m1, *m2;
  int arg;
  
  arg = 1;
  
  while (arg < argc) {
    if (strcmp (argv[arg], "-b") == 0) {
      if (arg + 1 >= argc) {
	usage ();
	return 1;
      }
      blob_dir = argv[++arg];
    } else if (strcmp (argv[arg], "-d") == 0) {
      if (arg + 1 >= argc) {
	usage ();
	return 1;
      }
      image_dir = argv[++arg];
    } else if (strcmp (argv[arg], "--delete") == 0) {
      do_delete = 1;
    } else {
      break;
    }
    arg++;
  }
		
  if (argc - arg != 2 || blob_dir == NULL || image_dir == NULL) {
    usage ();
    return 1;
  }
  
  m1 = parse_manifest (argv[arg++]);
  if (m1 == NULL)
    return 1;
  
  m2 = parse_manifest (argv[arg++]);
  if (m2 == NULL)
    return 1;

  blobs = NULL;
  if (do_delete) {
    blobs = string_collector_new ();
    manifest_diff (m1, m2, 1, blob_added, NULL, blob_changed, blobs);
  }
  
  /* We have to do this in two passes, one forward and one backwards.
     For the forward version we do additions (creating dirs before files in them)
     For the backwards version we do removals (remove files before the dirs the are in.
  */
  if (manifest_diff (m1, m2, 0, file1_added, file1_removed, file1_changed, NULL))
    manifest_diff (m1, m2, 1, file2_added, file2_removed, file2_changed, NULL);


  if (blobs)
    string_collector_free (blobs);
  
  return 0;
}
