#include "sha1.h"
#include "utils.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

int
safe_strcmp (const char *str1, const char *str2)
{
  if (str1 == NULL && str2 == NULL)
    return 0;

  if (str1 == NULL)
    return -1;
  if (str2 == NULL)
    return 1;

  return strcmp (str1, str2);
}

char *
read_symlink (char *path)
{
  char *buffer;
  size_t size;
  ssize_t res;

  size = 128;

  while (1) {
    buffer = malloc (size);
    res = readlink (path, buffer, size-1);
    if (res == -1) {
      free (buffer);
      buffer = NULL;
      break;
    }
    
    if (res != size - 1) {
      buffer[res] = 0;
      break;
    }
    
    free (buffer);
    size = size * 2;
  }
  
  return buffer;
}

static char *
sha1_to_ascii (unsigned char hash[20])
{
  const char *hex_digits = "0123456789abcdef";
  char *ascii_hash;
  int i;
  
  ascii_hash = malloc (41);
  for (i = 0; i < 20; i++)
    {
      ascii_hash[2*i] = hex_digits[hash[i] >> 4];
      ascii_hash[2*i+1] = hex_digits[hash[i] & 0xf];
    }
  ascii_hash[2*i] = 0;

  return ascii_hash;
}

char *
sha1_file (char *path)
{
  SHA_CTX ctx;
  unsigned char hash[20];
  FILE *file;
  char buffer[4096];
  size_t res;
  
  SHA1_Init(&ctx);
  file = fopen(path, "r");
  if (file)
    {
      while ((res = fread(buffer, 1, sizeof(buffer), file)) > 0)
	SHA1_Update(&ctx, buffer, res);

      fclose (file);
    }
  
  SHA1_Final(hash, &ctx);

  return sha1_to_ascii (hash);
}

int
copy_fd (int source_fd, char *dest, int mode)
{
  int fd_out;
  char buffer[8192];
  ssize_t res;
  size_t to_write;
  int pos;
  
  fd_out = open (dest, O_CREAT | O_WRONLY, mode);
  if (fd_out == -1) {
    return 0;
  }

  while (1) {
    res = read (source_fd, buffer, sizeof(buffer));
    if (res == -1) {
      if (errno == EINTR)
	continue;
      close (fd_out);
      return 0;
    }
    if (res == 0)
      break;

    to_write = res;
    pos = 0;

    while (pos < to_write) {
      res = write (fd_out, &buffer[pos], to_write - pos);
      if (res == -1) {
	if (errno == EINTR)
	  continue;
	close (fd_out);
	return 0;
      }
      if (res == 0)
	{
	  /* Short write?? */
	  close (fd_out);
	  return 0;
	}
      pos += res;
    }
  }
  
  close (fd_out);
  
  return 1;
}


int
copy_file (char *source, char *dest, int mode)
{
  int fd_in;
  int res;
  
  fd_in = open (source, O_RDONLY);
  if (fd_in == -1)
    return 0;

  res = copy_fd (fd_in, dest, mode);
  
  close(fd_in);

  return res;
}

char *
build_path (const char *dir, const char *file)
{
  char *path;

  path = malloc (strlen (dir) + 1 + strlen (file) + 1);
  strcpy (path, dir);
  strcat (path, "/");
  strcat (path, file);

  return path;
}


string_collector *
string_collector_new (void)
{
  string_collector *l;

  l = malloc (sizeof (string_collector));
  l->n_strings = 0;
  l->allocated_strings = 32;
  l->strings = malloc (sizeof (string_info) * l->allocated_strings);

  return l;
}

void
string_collector_free (string_collector *l)
{
  int i;

  for (i = 0; i < l->n_strings; i++)
    free (l->strings[i].string);
  free (l->strings);
  free (l);
}

static int
bsearch_strcmp (const void *_a, const void *_b)
{
  const char *a = _a;
  const string_info *b = _b;
  return strcmp (a, b->string);
}

string_info *
string_collector_lookup (string_collector *c,
			 const char *string)
{
  string_info *info;

  return bsearch (string, c->strings, c->n_strings, sizeof (string_info), bsearch_strcmp);
}

void
string_collector_add (string_collector *c,
		      const char *string)
{
  int start, end, mid;
  string_info *info;

  info = string_collector_lookup (c, string);
  if (info) {
    info->count++;
    return;
  }

  if (c->n_strings == c->allocated_strings) {
    c->allocated_strings = 2 * c->allocated_strings;
    c->strings = realloc (c->strings, sizeof (string_info) * c->allocated_strings);
  }

  start = 0;
  end = c->n_strings;

  while (start != end)
    {
      mid = start + (end - start) / 2;

      if (strcmp (string, c->strings[mid].string) < 0)
	end = mid;
      else if (strcmp (string, c->strings[mid].string) > 0)
	start = mid + 1;
    }

  memmove (&c->strings[start+1], &c->strings[start], (c->n_strings - start) * sizeof (string_info));
  c->strings[start].string = strdup (string);
  c->strings[start].count = 1;
  c->n_strings++;
}
