#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>

#include "utils.h"
#include "manifest.h"

static void
usage (void)
{
  fprintf (stderr, "Usage: verify-manifest <manifest> <image dir>\n");
}

int
verify_info (FileInfo *info, char *image_dir)
{
  struct stat buf;
  int res;
  FileType type;
  char *symlink, *sha1, *path;
  
  path = build_path (image_dir, info->name);

  if (lstat (path, &buf) != 0) {
    fprintf (stderr, "Unable to find file %s from manifest\n", info->name);
    free (path);
    return 0;
  }

  res = 1;
  
  if (info->type != FILE_TYPE_LINK) /* Symlinks have no own metadata */ {
    if (info->mode != (buf.st_mode & ~S_IFMT)) {
      res = 0;
      fprintf (stderr, "Wrong mode on file %s from manifest\n", info->name);
    }

    if (info->uid != buf.st_uid) {
      res = 0;
      fprintf (stderr, "Wrong uid on file %s from manifest\n", info->name);
    }

    if (info->gid != buf.st_gid) {
      res = 0;
      fprintf (stderr, "Wrong gid on file %s from manifest\n", info->name);
    }
  }

  type = get_file_type_from_stat (&buf);
  
  if (info->type != type) {
      fprintf (stderr, "Wrong type of file %s from manifest\n", info->name);
      res = 0;
  } else {
    if (type == FILE_TYPE_BLOB) {
      if (buf.st_size == 0) {
	if (strcmp (info->sha1, "empty") != 0) {
	  fprintf (stderr, "File %s from manifest empty when it shouldn't\n", info->name);
	  res = 0;
	}
      } else {
	if (strcmp (info->sha1, "empty") == 0) {
	  fprintf (stderr, "Nonempty file %s from manifest is empty\n", info->name);
	  res = 0;
	} else {
	  sha1 = sha1_file (path);
	  if (strcmp (info->sha1, sha1) != 0) {
	    fprintf (stderr, "Wrong sha1 hash for file %s from manifest\n", info->name);
	    res = 0;
	  }
	  free (sha1);
	}
      }
    } else if (type == FILE_TYPE_CHAR ||
	       type == FILE_TYPE_BLOCK) {
      if (info->dev != buf.st_rdev) {
	fprintf (stderr, "Wrong device node on file %s from manifest\n", info->name);
	res = 0;
      }
    } else if (type == FILE_TYPE_LINK) {
      symlink = read_symlink (path);
      if (symlink == NULL || strcmp (symlink, info->symlink) != 0) {
	fprintf (stderr, "Wrong symlink %s from manifest\n", info->name);
	res = 0;
      }
      free(symlink);
    }
  }

  free (path);

  return res;
}

int
main(int argc, char *argv[])
{
  Manifest *m;
  int arg;
  int do_blobs = 0;
  int do_blob_changes = 0;
  int i;
  int all_ok;
  char *manifest_path, *image_dir;
  
  arg = 1;

  if (argc - arg != 2) {
    usage ();
    return 1;
  }

  manifest_path = argv[arg++];
  m = parse_manifest (manifest_path);
  if (m == NULL) {
    fprintf (stderr, "Unable to parse manifest %s\n", manifest_path);
    return 1;
  }

  image_dir = argv[arg++];
  
  all_ok = 1;
  
  for (i = 0; i < m->n_infos; i++) {
    if (!verify_info (&m->infos[i], image_dir))
      all_ok = 0;
  }

  if (all_ok) {
    fprintf (stderr, "Image ok\n");
    return 0;
  } else 
    return 1;
}
